import lightning

def test_model_trainer_fit(multimodal_model, sample_train_val_datamodule):
    """Test end-to-end training."""
    # Get a sample batch for testing
    batch = next(iter(sample_train_val_datamodule.train_dataloader()))

    # Run a forward pass to verify the model works with the data
    y = multimodal_model(batch)

    # Train the model for one batch
    trainer = lightning.pytorch.trainer.trainer.Trainer(fast_dev_run=True, accelerator="cpu")
    trainer.fit(model=multimodal_model, datamodule=sample_train_val_datamodule)



"""Basic layers for composite models"""

import warnings

import torch
from torch import _VF, nn


class ImageEmbedding(nn.Module):
    """A embedding layer which concatenates an ID embedding as a new channel onto 3D inputs."""

    def __init__(self, num_embeddings, sequence_length, image_size_pixels, **kwargs):
        """A embedding layer which concatenates an ID embedding as a new channel onto 3D inputs.

        The embedding is a single 2D image and is appended at each step in the 1st dimension
        (assumed to be time).

        Args:
            num_embeddings: Size of the dictionary of embeddings
            sequence_length: The time sequence length of the data.
            image_size_pixels: The spatial size of the image. Assumed square.
            **kwargs: See `torch.nn.Embedding` for more possible arguments.
        """
        super().__init__()
        self.image_size_pixels = image_size_pixels
        self.sequence_length = sequence_length
        self._embed = nn.Embedding(
            num_embeddings=num_embeddings,
            embedding_dim=image_size_pixels * image_size_pixels,
            **kwargs,
        )

    def forward(self, x, id):
        """Append ID embedding to image"""
        emb = self._embed(id)
        emb = emb.reshape((-1, 1, 1, self.image_size_pixels, self.image_size_pixels))
        emb = emb.repeat(1, 1, self.sequence_length, 1, 1)
        x = torch.cat((x, emb), dim=1)
        return x


class CompleteDropoutNd(nn.Module):
    """A layer used to completely drop out all elements of a N-dimensional sample.

    Each sample will be zeroed out independently on every forward call with probability `p` using
    samples from a Bernoulli distribution.

    """

    __constants__ = ["p", "inplace", "n_dim"]
    p: float
    inplace: bool
    n_dim: int

    def __init__(self, n_dim, p=0.5, inplace=False):
        """A layer used to completely drop out all elements of a N-dimensional sample.

        Args:
            n_dim: Number of dimensions of each sample not including channels. E.g. a sample with
                shape (channel, time, height, width) would use `n_dim=3`.
            p: probability of a channel to be zeroed. Default: 0.5
            training: apply dropout if is `True`. Default: `True`
            inplace: If set to `True`, will do this operation in-place. Default: `False`
        """
        super().__init__()
        if p < 0 or p > 1:
            raise ValueError(
                "dropout probability has to be between 0 and 1, " "but got {}".format(p)
            )
        self.p = p
        self.inplace = inplace
        self.n_dim = n_dim

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Run dropout"""
        p = self.p
        inp_dim = input.dim()

        if inp_dim not in (self.n_dim + 1, self.n_dim + 2):
            warn_msg = (
                f"CompleteDropoutNd: Received a {inp_dim}-D input. Expected either a single sample"
                f" with {self.n_dim+1} dimensions, or a batch of samples with {self.n_dim+2}"
                " dimensions."
            )
            warnings.warn(warn_msg)

        is_batched = inp_dim == self.n_dim + 2
        if not is_batched:
            input = input.unsqueeze_(0) if self.inplace else input.unsqueeze(0)

        input = input.unsqueeze_(1) if self.inplace else input.unsqueeze(1)

        result = (
            _VF.feature_dropout_(input, p, self.training)
            if self.inplace
            else _VF.feature_dropout(input, p, self.training)
        )

        result = result.squeeze_(1) if self.inplace else result.squeeze(1)

        if not is_batched:
            result = result.squeeze_(0) if self.inplace else result.squeeze(0)

        return result



""" Data module for pytorch lightning """

from ocf_data_sampler.torch_datasets.datasets.site import SitesDataset
from ocf_data_sampler.torch_datasets.sample.site import SiteSample
from torch.utils.data import Dataset

from pvnet.data.base_datamodule import BaseDataModule, PremadeSamplesDataset


class SiteDataModule(BaseDataModule):
    """Datamodule for training pvnet and using pvnet pipeline in `ocf-data-sampler`."""

    def __init__(
        self,
        configuration: str | None = None,
        sample_dir: str | None = None,
        batch_size: int = 16,
        num_workers: int = 0,
        prefetch_factor: int | None = None,
        train_period: list[str | None] = [None, None],
        val_period: list[str | None] = [None, None],
    ):
        """Datamodule for training pvnet architecture.

        Can also be used with pre-made batches if `sample_dir` is set.

        Args:
            configuration: Path to configuration file.
            sample_dir: Path to the directory of pre-saved samples. Cannot be used together with
                `configuration` or '[train/val]_period'.
            batch_size: Batch size.
            num_workers: Number of workers to use in multiprocess batch loading.
            prefetch_factor: Number of data will be prefetched at the end of each worker process.
            train_period: Date range filter for train dataloader.
            val_period: Date range filter for val dataloader.

        """
        super().__init__(
            configuration=configuration,
            sample_dir=sample_dir,
            batch_size=batch_size,
            num_workers=num_workers,
            prefetch_factor=prefetch_factor,
            train_period=train_period,
            val_period=val_period,
        )

    def _get_streamed_samples_dataset(self, start_time, end_time) -> Dataset:
        return SitesDataset(self.configuration, start_time=start_time, end_time=end_time)

    def _get_premade_samples_dataset(self, subdir) -> Dataset:
        split_dir = f"{self.sample_dir}/{subdir}"
        return PremadeSamplesDataset(split_dir, SiteSample)



"""Basic blocks for PV-site encoders"""
from abc import ABCMeta, abstractmethod

from torch import nn


class AbstractSitesEncoder(nn.Module, metaclass=ABCMeta):
    """Abstract class for encoder for output data from multiple PV sites.

    The encoder will take an input of shape (batch_size, sequence_length, num_sites)
    and return an output of shape (batch_size, out_features).
    """

    def __init__(
        self,
        sequence_length: int,
        num_sites: int,
        out_features: int,
    ):
        """Abstract class for PV site-level encoder.

        Args:
            sequence_length: The time sequence length of the data.
            num_sites: Number of PV sites in the input data.
            out_features: Number of output features.
        """
        super().__init__()
        self.sequence_length = sequence_length
        self.num_sites = num_sites
        self.out_features = out_features

    @abstractmethod
    def forward(self):
        """Run model forward"""
        pass
